import { DeepRequired, MarkOptional, Merge } from 'ts-essentials';
import { TornadoInstance, TornadoProxy } from './deth';
import { RelayerProperties as RelayerDataProperties } from '@tornado/sdk-data';
import { DepositInfo } from '@tornado/sdk-crypto';
import { TransactionRequest } from '@ethersproject/abstract-provider';
import { BigNumber, providers } from 'ethers';
import { AsyncUtils } from '@tornado/sdk-utils';
import { Docs, Cache, Keys, Options as DataOptions } from '@tornado/sdk-data';
import { Chain, Synchronizer, Options as ChainOptions } from '@tornado/sdk-chain';
export declare namespace Options {
    namespace Core {
        interface Deposit {
            depositsPerInstance?: Array<number>;
            doNotPopulate?: boolean;
        }
        type Invoice = Deposit;
        interface BuildDepositProof {
            gasPrice?: BigNumber;
            gasPriceCushion?: BigNumber;
            tokenDecimals?: number;
            ethPurchaseAmounts?: Array<BigNumber>;
            checkNotesSpent?: boolean;
            checkKnownRoot?: boolean;
            merkleTreeHeight?: number;
        }
    }
    type Sync = ChainOptions.Sync;
    type Cache = DataOptions.Cache;
}
export declare namespace Transactions {
    interface Deposit {
        request: TransactionRequest;
        invoice?: string;
        note?: string;
    }
    type Invoice = Deposit;
}
export declare class DepositCache extends Cache.Syncable<Docs.Deposit> {
    buildDoc(response: any): Docs.Deposit;
    getErrorHandlers(): Array<AsyncUtils.ErrorHandler>;
    getCallbacks(instance: TornadoInstance): Array<AsyncUtils.Callback>;
}
export declare class WithdrawalCache extends Cache.Syncable<Docs.Withdrawal> {
    buildDoc(response: any): Docs.Withdrawal;
    getErrorHandlers(): Array<AsyncUtils.ErrorHandler>;
    getCallbacks(instance: TornadoInstance): Array<AsyncUtils.Callback>;
}
type Provider = providers.Provider;
type RelayerProperties = MarkOptional<Pick<RelayerDataProperties, 'address' | 'serviceFeePercent' | 'prices'>, 'serviceFeePercent' | 'prices'>;
export declare class Core extends Synchronizer {
    private _mutex;
    private _chain?;
    private _notes?;
    private _invoices?;
    get chain(): Chain;
    constructor();
    private _checkProvider;
    connect(provider: Provider): Promise<void>;
    getInstances(keys: Array<{
        token: string;
        denomination: number | string;
    }>): Array<TornadoInstance>;
    getInstance(token: string, denomination: number | string): TornadoInstance;
    getProxy(): TornadoProxy;
    createDepositProof(instance: TornadoInstance, relayerProperties: RelayerProperties, recipientAddress: string, depositInfo: DepositInfo, options?: Options.Core.BuildDepositProof): Promise<Array<string>>;
    /**
     * @param instance This is the Tornado Instance which will be withdrawn from.
     * @param relayerProperties The properties of the relayer that is going to be used for the withdrawals. If the service fee is 0, it is assumed that there is no relayer, but that a manual wallet withdraw is being made. These properties are included in the ZK proof.
     * @param recipientAddresses The recipient addresses which should receive the withdrawals, in order.
     * @param depositInfo These represent the public and private values, reconstructed from the deposit note, generated during the building of deposit transactions, used for building the proof of knowledge statement for withdrawal, for each withdrawal (in this context).
     * @param options Numerous options which most importantly allow a user to specify whether he is buying ETH, whether to check proof data validity and finally to modulate the gas prices which will be used to calculate the gas fees paid to the relayer.
     * @returns The proofs for which the user should then decide whether to use a relayer (recommended, but decide carefully which one) or use his own wallet (if needed).
     */
    createDepositProofs(instance: TornadoInstance, relayerProperties: RelayerProperties, recipientAddresses: Array<string>, depositInfo: Array<DepositInfo>, options?: Options.Core.BuildDepositProof): Promise<Array<Array<string>>>;
    private _calcWithdrawalFee;
    /**
     * @param instanceName The name of the instance as created in `_sync` function.
     * @param commitments The commitments for which the leaf index values are to be noted down extra.
     * @returns The result of concatenating the array of leaf indices found by matching them with the provided commitment values, followed by the array of all leaf indices, including all of the formerly mentioned values given that they are valid. Values which have not been matched, meaning probably invalid values, will be `0`.
     */
    private _findLeavesAndIndices;
    loadNotes(indexes?: Array<number>, keys?: Partial<Keys.InstanceLookup>, options?: Options.Cache): Promise<Array<DepositInfo>>;
    parseNotes(notes: Array<string>): Array<DepositInfo>;
    parseNote(note: string): DepositInfo;
    clearListener(instance: TornadoInstance | string, event?: Function | number, listenerIndex?: number): void;
    clearListeners(instance: TornadoInstance | string): void;
    listenForDeposits(instance: TornadoInstance | string): void;
    listenForWithdrawals(instance: TornadoInstance | string): void;
    listenForInstanceEvents(instance: TornadoInstance | string, event?: Function | number): void;
    private _instanceEventToFilter;
    private _resolveInstanceEvent;
    private _resolveInstance;
    /**
     * This is the main function to build a single Tornado Cash Classic deposit. An address need not be supplied because the returned note proves a deposit.
     * @param instance The TornadoInstance for which to build transactions.
     * @param options Whether or not to populate the transactions (only in the sense of encoding transaction data), and whether to backup notes and invoices. Defaults: `depositsPerInstance = [1], doNotPopulate = false, backup { notes = true, invoices = false }` Deposits per instance are hardcoded to 1, since we're doing a single transaction.
     * @returns A promise which resolves to the created transaction.
     */
    createDepositTransaction(instance: TornadoInstance, options?: Options.Core.Deposit): Transactions.Deposit;
    /**
     * This is the main function which is used to build Tornado Cash Classic deposit transactions. An address need not be supplied because the returned note proves a deposit.
     * @param instances The TornadoInstance instances for which to build transactions.
     * @param options The number of deposits per instance, whether or not to populate the transactions (only in the sense of encoding transaction data), and whether to backup notes and invoices. Defaults: `depositsPerInstance = [1]*instance_num, doNotPopulate = false, backup { notes = true, invoices = false }`
     * @returns A promise which resolves to the created transactions.
     * @todo TODO: Maybe this should be sync and deposit backups should be async somewhere else
     */
    createDepositTransactions(instances: Array<TornadoInstance>, options?: Options.Core.Deposit): Array<Transactions.Deposit>;
    syncDeposits(instance: TornadoInstance, options?: Merge<Options.Sync, Options.Cache>): Promise<void>;
    syncWithdrawals(instance: TornadoInstance, options?: Merge<Options.Sync, Options.Cache>): Promise<void>;
    protected _populateSyncOptions(options: Options.Sync): Promise<DeepRequired<Options.Sync>>;
    backupNote(instance: TornadoInstance, transaction: Transactions.Deposit, options?: Options.Cache): Promise<void>;
    backupInvoice(instance: TornadoInstance, transaction: Transactions.Deposit, options?: Options.Cache): Promise<void>;
    backupNotes(instance: TornadoInstance, transactions: Array<Transactions.Deposit>, options?: Options.Cache): Promise<void>;
    backupInvoices(instance: TornadoInstance, transactions: Array<Transactions.Deposit>, options?: Options.Cache): Promise<void>;
    private _backupDepositData;
    exportCacheZip(cacheName: string, outDirPath?: string, close?: boolean, options?: Options.Cache): Promise<void>;
    exportCacheJson(cacheName: string, outDirPath?: string, close?: boolean, options?: Options.Cache): Promise<void>;
    loadNotesCache(options?: Options.Cache): Cache.Base<Docs.Note>;
    loadInvoicesCache(options?: Options.Cache): Cache.Base<Docs.Invoice>;
    loadDepositCache(name: string, options?: Options.Cache): DepositCache;
    loadWithdrawalCache(name: string, options?: Options.Cache): WithdrawalCache;
    loadCache<D extends Docs.Base, C extends Cache.Syncable<D> = Cache.Syncable<D>>(name: string, options?: Options.Cache): C;
    loadInstance(chainId: number | string, token: string, denomination: number | string): TornadoInstance;
}
export {};
//# sourceMappingURL=index.d.ts.map