"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Core = exports.WithdrawalCache = exports.DepositCache = void 0;
const ethers_1 = require("ethers");
const utils_1 = require("ethers/lib/utils");
const snarkjs_1 = require("snarkjs");
// @ts-ignore
const pouchdb_collate_1 = require("pouchdb-collate");
// Local imports
const sdk_crypto_1 = require("@tornado/sdk-crypto");
const sdk_utils_1 = require("@tornado/sdk-utils");
const sdk_data_1 = require("@tornado/sdk-data");
const sdk_chain_1 = require("@tornado/sdk-chain");
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FOR SYNCHRONIZATION ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class DepositCache extends sdk_data_1.Cache.Syncable {
    buildDoc(response) {
        return new sdk_data_1.Docs.Deposit(response);
    }
    getErrorHandlers() {
        return [sdk_chain_1.syncErrorHandler];
    }
    getCallbacks(instance) {
        return [
            (fromBlock, toBlock) => {
                return instance.queryFilter(instance.filters.Deposit(null, null, null), fromBlock, toBlock);
            }
        ];
    }
}
exports.DepositCache = DepositCache;
class WithdrawalCache extends sdk_data_1.Cache.Syncable {
    buildDoc(response) {
        return new sdk_data_1.Docs.Withdrawal(response);
    }
    getErrorHandlers() {
        return [sdk_chain_1.syncErrorHandler];
    }
    getCallbacks(instance) {
        return [
            (fromBlock, toBlock) => {
                return instance.queryFilter(instance.filters.Withdrawal(null, null, null, null), fromBlock, toBlock);
            }
        ];
    }
}
exports.WithdrawalCache = WithdrawalCache;
class Core extends sdk_chain_1.Synchronizer {
    _mutex;
    _chain;
    _notes;
    _invoices;
    get chain() {
        this._checkProvider('chain');
        return this._chain;
    }
    constructor() {
        super();
        this._mutex = new sdk_utils_1.AsyncUtils.SimpleMutex();
    }
    _checkProvider(parentCallName) {
        if (!this._chain)
            throw sdk_utils_1.ErrorUtils.getError('Core.' + parentCallName + ': you must first connect a provider!');
    }
    async connect(provider) {
        if (!this._chain)
            this._chain = new sdk_chain_1.Chain(provider);
        else
            this._chain.provider = provider;
        await this._chain.fetchChainData();
    }
    getInstances(keys) {
        this._checkProvider('getInstances');
        return keys.map((key) => sdk_chain_1.Contracts.getInstance(String(this.chain.id), key.token, String(key.denomination), this.chain.provider));
    }
    getInstance(token, denomination) {
        this._checkProvider('getInstance');
        return this.loadInstance(this.chain.id, token, denomination);
    }
    getProxy() {
        this._checkProvider('getProxy');
        return sdk_chain_1.Contracts.getProxy(String(this.chain.id), this.chain.provider);
    }
    async createDepositProof(instance, relayerProperties, recipientAddress, depositInfo, options) {
        return (await this.createDepositProofs(instance, relayerProperties, [recipientAddress], [depositInfo], options))[0];
    }
    // TODO: Abstract out verification parts of this and provide it as a standalone service in crypto or somewhere else
    /**
     * @param instance This is the Tornado Instance which will be withdrawn from.
     * @param relayerProperties The properties of the relayer that is going to be used for the withdrawals. If the service fee is 0, it is assumed that there is no relayer, but that a manual wallet withdraw is being made. These properties are included in the ZK proof.
     * @param recipientAddresses The recipient addresses which should receive the withdrawals, in order.
     * @param depositInfo These represent the public and private values, reconstructed from the deposit note, generated during the building of deposit transactions, used for building the proof of knowledge statement for withdrawal, for each withdrawal (in this context).
     * @param options Numerous options which most importantly allow a user to specify whether he is buying ETH, whether to check proof data validity and finally to modulate the gas prices which will be used to calculate the gas fees paid to the relayer.
     * @returns The proofs for which the user should then decide whether to use a relayer (recommended, but decide carefully which one) or use his own wallet (if needed).
     */
    async createDepositProofs(instance, relayerProperties, recipientAddresses, depositInfo, options) {
        this._checkProvider('createDepositProofs');
        // Extract commitments and nullifier hashes
        const hexCommitments = [];
        const hexNullifierHashes = [];
        const purchaseAmounts = options?.ethPurchaseAmounts ?? new Array(depositInfo.length).fill(ethers_1.BigNumber.from(0));
        if (depositInfo.length !== recipientAddresses.length)
            throw sdk_utils_1.ErrorUtils.getError('Core.createDepositProofs: the number of recipients must equal the length of depositInfo.');
        if (depositInfo.length !== purchaseAmounts.length)
            throw sdk_utils_1.ErrorUtils.getError('Core.createDepositProofs: if purchase amounts is specified, it must equal the length of depositInfo.');
        depositInfo.forEach((deposit) => {
            hexCommitments.push(deposit.hexCommitment);
            hexNullifierHashes.push(deposit.hexNullifierHash);
        });
        // Determine cache name
        const { network, token, denomination } = await sdk_data_1.Onchain.getInstanceLookupKeys(instance.address);
        const name = 'Deposits' + (network + token + denomination).toUpperCase();
        // Find all leaves & indices by reading from cache
        const [leaves, leafIndices] = await this._findLeavesAndIndices(name, hexCommitments);
        const invalidCommitments = [];
        this.emit('debug', `\nFound leaves and indices, num leaves: ${leaves.length}, indices: [${leafIndices.join(', ')}]`);
        // Determine whether we will be checking whether notes are spent
        const checkSpent = options?.checkNotesSpent !== false;
        const spentNotes = [];
        this.emit('debug', `\nCheck spent notes? => ${checkSpent}`);
        // If yes, immediately check it with the supplied Tornado Instance
        const checkSpentArray = checkSpent ? await instance.isSpentArray(hexNullifierHashes) : undefined;
        if (checkSpent)
            this.emit('debug', `\nSpent array: [${checkSpentArray?.join(', ')}]`);
        // Check whether a commitment has not been found in all deposits, meaning that it is invalid
        // Also add the invalid commitments. We can do leafIndices[i] because the matched one are concatenated
        // at the start
        for (let i = 0, len = depositInfo.length; i < len; i++) {
            if (!leafIndices[i])
                invalidCommitments.push(hexCommitments[i]);
            if (checkSpent && checkSpentArray[i])
                spentNotes.push(hexNullifierHashes[i]);
        }
        // If something is wrong, throw
        const commitmentsAreInvalid = invalidCommitments.length !== 0;
        const notesAreSpent = spentNotes.length !== 0;
        if (commitmentsAreInvalid || notesAreSpent)
            throw sdk_utils_1.ErrorUtils.getError(`Core.createDepositProofs: ` +
                (commitmentsAreInvalid
                    ? `following commitments are invalid:\n\n${invalidCommitments.join('\n')}\n\n`
                    : '') +
                (notesAreSpent
                    ? `${commitmentsAreInvalid ? 'and ' : ''}following notes are already spent or invalid:\n\n${spentNotes.join('\n')}\n\n`
                    : ''));
        // Otherwise, build the merkle tree from the leaves
        const merkleTree = sdk_crypto_1.Primitives.buildMerkleTree({
            height: options?.merkleTreeHeight ?? sdk_data_1.Constants.MERKLE_TREE_HEIGHT,
            leaves: leaves
        });
        const root = ethers_1.BigNumber.from(merkleTree.root).toHexString();
        const checkKnownRoot = options?.checkKnownRoot ?? true;
        this.emit('debug', `\nMerkle root: ${root}, check known? => ${checkKnownRoot}`);
        // Check whether the root is valid
        if (checkKnownRoot && !(await instance.isKnownRoot(root)))
            throw sdk_utils_1.ErrorUtils.getError('Core.createDepositProofs: the merkle tree created is not valid, something went wrong with syncing.');
        // Rest of note invariant arguments
        const inputsForProofs = [];
        const gasPrice = options?.gasPrice ?? (await this.chain.getGasPrice());
        const gasPriceCushion = options?.gasPrice ?? gasPrice.mul(10).div(100);
        // In reality, if a manual withdraw is made, we don't differentiate it from a relayer withdraw
        // Since it is only serviceFee 0 AND without a token price, the function will not buy more tokens
        const serviceFeePercent = relayerProperties.serviceFeePercent ?? 0;
        const tokenPrice = relayerProperties.prices?.get(token);
        const decimals = ethers_1.BigNumber.from(10).pow(options?.tokenDecimals ?? (await sdk_data_1.Onchain.getTokenDecimals(network, token)));
        const toWithdraw = ethers_1.BigNumber.from(+denomination * 10 ** denomination.length)
            .mul(decimals)
            .div(10 ** denomination.length);
        const native = token == this.chain.symbol;
        if (!tokenPrice && !native)
            throw sdk_utils_1.ErrorUtils.getError('Core.createDepositProofs: a token price MUST be supplied if the token withdrawn is not native.');
        this.emit('debug', `\nProof building, invariant data: [${[
            gasPrice.toString(),
            gasPriceCushion.toString(),
            serviceFeePercent,
            tokenPrice,
            decimals.toString(),
            toWithdraw.toString()
        ].join(', ')}]\n`);
        // Compute proofs
        for (let i = 0, len = depositInfo.length; i < len; i++) {
            inputsForProofs.push({
                public: {
                    root: root,
                    tree: merkleTree,
                    leafIndex: leafIndices[i],
                    hexNullifierHash: depositInfo[i].hexNullifierHash,
                    recipientAddress: recipientAddresses[i],
                    relayerAddress: relayerProperties.address,
                    fee: this._calcWithdrawalFee(toWithdraw, decimals, gasPrice, gasPriceCushion, serviceFeePercent, purchaseAmounts[i], 
                    // This is our flag whether it's a token or not
                    native ? undefined : tokenPrice),
                    // @ts-expect-error
                    refund: purchaseAmounts[i] ? (0, snarkjs_1.bigInt)(purchaseAmounts[i].toString()) : (0, snarkjs_1.bigInt)(0)
                },
                private: {
                    nullifier: depositInfo[i].nullifier,
                    secret: depositInfo[i].secret
                }
            });
        }
        return await sdk_crypto_1.Primitives.calcDepositProofs(inputsForProofs);
    }
    _calcWithdrawalFee(toWithdraw, decimals, gasPrice, gasPriceCushion, relayerServiceFee, ethBought, tokenPriceInEth) {
        const factor = 10 ** String(relayerServiceFee).length;
        const baseRelayerFee = toWithdraw.mul(ethers_1.BigNumber.from(relayerServiceFee * factor)).div(factor);
        const txCost = gasPrice.add(gasPriceCushion).mul(5e5);
        if (tokenPriceInEth) {
            // @ts-expect-error
            return (0, snarkjs_1.bigInt)(txCost.add(ethBought).mul(decimals).div(tokenPriceInEth).add(baseRelayerFee).toString());
        }
        // @ts-expect-error
        else
            return (0, snarkjs_1.bigInt)(txCost.add(baseRelayerFee).toString());
    }
    /**
     * @param instanceName The name of the instance as created in `_sync` function.
     * @param commitments The commitments for which the leaf index values are to be noted down extra.
     * @returns The result of concatenating the array of leaf indices found by matching them with the provided commitment values, followed by the array of all leaf indices, including all of the formerly mentioned values given that they are valid. Values which have not been matched, meaning probably invalid values, will be `0`.
     */
    async _findLeavesAndIndices(instanceName, commitments) {
        const indices = new Array(commitments.length).fill(0);
        const leaves = [];
        const cache = this.loadCache(instanceName);
        // Shallow copy so we can find indexes again for commitments
        const commitmentsCopy = [...commitments];
        const docs = await cache.db.allDocs();
        // If no docs in cache throw and stop
        if (docs.total_rows === 0) {
            await cache.clear();
            throw sdk_utils_1.ErrorUtils.getError(`Core.buildMerkleTree: events for instance ${instanceName} have not been synchronized.`);
        }
        // Otherwise start looking for commitment leaf indices and also pick up all other leafs on the way
        for (const row of docs.rows) {
            let index = -1;
            const [, leafIndex, loadedCommitment] = (0, pouchdb_collate_1.parseIndexableString)(row.id);
            // Search only if there is some left
            if (commitments.length !== 0)
                index = commitments.findIndex((commitment) => commitment === loadedCommitment);
            // If some commitment is found then add the leaf index
            if (index !== -1) {
                // Add it there where we intended for it to be originally
                indices[commitmentsCopy.findIndex((commitment) => commitment === loadedCommitment)] = leafIndex;
                commitments.splice(index, 1);
                this.emit('debug', `\nMatched commitment ${loadedCommitment} @ leaf index ${leafIndex}, leftover commitments:\n\n${commitments.join('\n')}\n`);
            }
            // In any case push every leaf
            leaves.push(ethers_1.BigNumber.from(loadedCommitment).toString());
        }
        // Concat matched and all leaf indices
        return [leaves, indices];
    }
    async loadNotes(indexes, keys, options) {
        const cache = this._notes ?? new sdk_data_1.Cache.Base('DepositNotes', options);
        const rows = await cache.getRows();
        let docs = [];
        let notes = [];
        if (indexes)
            for (let i = 0, len = rows.length; i < len; i++) {
                docs.push(rows[indexes[i]].doc);
            }
        else
            docs = rows.map((row) => row.doc);
        if (keys)
            docs.forEach((doc) => {
                const idNetworkMatches = doc && (keys.network ? keys.network === doc?.network : true);
                const andTokenSymbolMatches = idNetworkMatches && (keys.token ? keys.token === doc?.token : true);
                const lastlyDenominationMatches = andTokenSymbolMatches && (keys.denomination ? keys.denomination === doc?.denomination : true);
                if (lastlyDenominationMatches && doc?.note)
                    notes.push(doc.note);
            });
        else
            notes = docs.filter((doc) => sdk_utils_1.ObjectUtils.exists(doc?.note)).map((doc) => doc.note);
        return this.parseNotes(notes);
    }
    parseNotes(notes) {
        return notes.map((note) => sdk_crypto_1.Primitives.parseNote(note));
    }
    parseNote(note) {
        return this.parseNotes([note])[0];
    }
    clearListener(instance, event = 0, listenerIndex = 0) {
        const _instance = this._resolveInstance(instance);
        const filter = this._resolveInstanceEvent(_instance, event);
        this.clearListenerByIndex(_instance, this._instanceEventToFilter(filter, _instance.filters.Deposit), listenerIndex);
    }
    clearListeners(instance) {
        this._resolveInstance(instance).removeAllListeners();
    }
    listenForDeposits(instance) {
        this.listenForInstanceEvents(instance, 0);
    }
    listenForWithdrawals(instance) {
        this.listenForInstanceEvents(instance, 1);
    }
    listenForInstanceEvents(instance, event = 0) {
        let _instance;
        let key;
        if (typeof instance !== 'string') {
            const { network, token, denomination } = sdk_data_1.Onchain.getInstanceLookupKeysSync(instance.address);
            _instance = this.loadInstance(network, token, denomination);
            key = network + token + denomination;
        }
        else {
            key = instance.toLowerCase();
            _instance = this._resolveInstance(key);
        }
        const filter = this._resolveInstanceEvent(_instance, event);
        const isDeposit = filter == _instance.filters.Deposit;
        const cache = isDeposit
            ? this.loadDepositCache('Deposits' + key.toUpperCase())
            : this.loadWithdrawalCache('Withdrawals' + key.toUpperCase());
        this.listenForEvents(isDeposit ? 'deposit' : 'withdrawal', _instance, this._instanceEventToFilter(filter, _instance.filters.Deposit), cache);
    }
    _instanceEventToFilter(event, depositEvent) {
        return event == depositEvent ? event(null, null, null) : event(null, null, null, null);
    }
    _resolveInstanceEvent(instance, event = 0) {
        let filter;
        if (typeof event === 'number') {
            filter = event === 0 ? instance.filters.Deposit : instance.filters.Withdrawal;
        }
        else
            filter = event;
        return filter;
    }
    _resolveInstance(instance) {
        let _instance;
        if (typeof instance === 'string') {
            instance = instance.toLowerCase();
            const regexp = /([0-9]+)([a-z]+)([0-9.]+)/;
            const matches = instance.match(regexp)?.slice(1);
            if (!matches || matches.length === 0)
                throw sdk_utils_1.ErrorUtils.getError('Core._resolveInstance: instance string key invalid.');
            _instance = this.loadInstance(matches[0], matches[1], matches[2]);
        }
        else
            _instance = instance;
        return _instance;
    }
    /**
     * This is the main function to build a single Tornado Cash Classic deposit. An address need not be supplied because the returned note proves a deposit.
     * @param instance The TornadoInstance for which to build transactions.
     * @param options Whether or not to populate the transactions (only in the sense of encoding transaction data), and whether to backup notes and invoices. Defaults: `depositsPerInstance = [1], doNotPopulate = false, backup { notes = true, invoices = false }` Deposits per instance are hardcoded to 1, since we're doing a single transaction.
     * @returns A promise which resolves to the created transaction.
     */
    createDepositTransaction(instance, options) {
        let opts = options ?? {};
        opts.depositsPerInstance = [1];
        return this.createDepositTransactions([instance], opts)[0];
    }
    /**
     * This is the main function which is used to build Tornado Cash Classic deposit transactions. An address need not be supplied because the returned note proves a deposit.
     * @param instances The TornadoInstance instances for which to build transactions.
     * @param options The number of deposits per instance, whether or not to populate the transactions (only in the sense of encoding transaction data), and whether to backup notes and invoices. Defaults: `depositsPerInstance = [1]*instance_num, doNotPopulate = false, backup { notes = true, invoices = false }`
     * @returns A promise which resolves to the created transactions.
     * @todo TODO: Maybe this should be sync and deposit backups should be async somewhere else
     */
    createDepositTransactions(instances, options) {
        this._checkProvider('createDepositTransactions');
        const depositsPerInstance = options?.depositsPerInstance ?? new Array(instances.length).fill(1);
        const doNotPopulate = options?.doNotPopulate ?? false;
        if (depositsPerInstance.length != instances.length)
            throw sdk_utils_1.ErrorUtils.getError('Core.createDepositTx: number of deposit amount elements must equal the number of instances!');
        const chainId = this.chain.id;
        const proxy = sdk_chain_1.Contracts.getProxy(String(chainId), this.chain.provider);
        const txs = [];
        for (let i = 0, nInstances = instances.length; i < nInstances; i++) {
            const { network, token, denomination } = sdk_data_1.Onchain.getInstanceLookupKeysSync(instances[i].address);
            const pathstring = network + token + denomination;
            for (let d = 0, nDeposits = depositsPerInstance[i]; d < nDeposits; d++) {
                const deposit = sdk_crypto_1.Primitives.createDeposit();
                const note = sdk_crypto_1.Primitives.createNote(deposit.preimage);
                if (!doNotPopulate) {
                    txs.push({
                        request: {
                            to: proxy.address,
                            data: proxy.interface.encodeFunctionData('deposit', [
                                instances[i].address,
                                deposit.hexCommitment,
                                []
                            ]),
                            value: token == 'eth' ? (0, utils_1.parseUnits)(denomination) : ethers_1.BigNumber.from(0)
                        },
                        note: pathstring + '_' + note,
                        invoice: pathstring + '_' + deposit.hexCommitment
                    });
                }
                else
                    txs.push({
                        request: {},
                        note: pathstring + '_' + note,
                        invoice: pathstring + '_' + deposit.hexCommitment
                    });
            }
        }
        return txs;
    }
    async syncDeposits(instance, options) {
        const { network, token, denomination } = await sdk_data_1.Onchain.getInstanceLookupKeys(instance.address);
        const pathstring = network + token + denomination;
        options = options ?? {};
        options.startBlock = await sdk_data_1.Onchain.getInstanceDeployBlockNum(network, token, denomination);
        const cache = this.loadDepositCache('Deposits' + pathstring.toUpperCase(), {
            dirPath: options.dirPath,
            adapter: options.adapter,
            persistent: options.persistent
        });
        await this.sync('deposit', instance.filters.Deposit(null, null, null), instance, cache, options);
    }
    async syncWithdrawals(instance, options) {
        const { network, token, denomination } = await sdk_data_1.Onchain.getInstanceLookupKeys(instance.address);
        const pathstring = network + token + denomination;
        options = options ?? {};
        options.startBlock = await sdk_data_1.Onchain.getInstanceDeployBlockNum(network, token, denomination);
        const cache = this.loadWithdrawalCache('Withdrawals' + pathstring.toUpperCase(), {
            dirPath: options.dirPath,
            adapter: options.adapter,
            persistent: options.persistent
        });
        await this.sync('withdrawal', instance.filters.Withdrawal(null, null, null), instance, cache, options);
    }
    async _populateSyncOptions(options) {
        options.targetBlock = options.targetBlock ?? (await this.chain.latestBlockNum());
        return super._populateSyncOptions(options);
    }
    async backupNote(instance, transaction, options) {
        await this.backupNotes(instance, [transaction], options);
    }
    async backupInvoice(instance, transaction, options) {
        await this.backupInvoices(instance, [transaction], options);
    }
    async backupNotes(instance, transactions, options) {
        let cache = this._notes ?? new sdk_data_1.Cache.Base('DepositNotes', options);
        await this._backupDepositData(instance.address, transactions, cache);
        if (!this._notes)
            this._notes = cache;
    }
    async backupInvoices(instance, transactions, options) {
        let cache = this._invoices ?? new sdk_data_1.Cache.Base('DepositInvoices', options);
        await this._backupDepositData(instance.address, transactions, cache);
        if (!this._invoices)
            this._invoices = cache;
    }
    async _backupDepositData(address, transactions, cache) {
        const { network, token, denomination } = await sdk_data_1.Onchain.getInstanceLookupKeys(address);
        const notes = cache.name.length === 12 ? true : false;
        const name = notes ? 'notes' : 'invoices';
        // We need a mutex here
        const release = await this._mutex.acquire(name);
        let id = +(await cache.db.info()).update_seq;
        await cache.db
            .bulkDocs(transactions.map((transaction) => {
            if (notes)
                return new sdk_data_1.Docs.Note(++id, network, token, denomination, transaction.note);
            else
                return new sdk_data_1.Docs.Invoice(++id, network, token, denomination, transaction.invoice);
        }))
            .catch((err) => {
            throw sdk_utils_1.ErrorUtils.ensureError(err);
        });
        // Release
        release();
    }
    async exportCacheZip(cacheName, outDirPath, close, options) {
        const cache = new sdk_data_1.Cache.Base(cacheName, options);
        await cache.zip(outDirPath, close);
        if (close === true)
            this.caches.delete(cacheName);
    }
    async exportCacheJson(cacheName, outDirPath, close, options) {
        const cache = new sdk_data_1.Cache.Base(cacheName, options);
        await cache.jsonify(outDirPath);
        if (close === true)
            this.caches.delete(cacheName);
    }
    loadNotesCache(options) {
        if (!this._notes)
            this._notes = new sdk_data_1.Cache.Base('DepositNotes', options);
        return this._notes;
    }
    loadInvoicesCache(options) {
        if (!this._invoices)
            this._invoices = new sdk_data_1.Cache.Base('DepositInvoices', options);
        return this._invoices;
    }
    loadDepositCache(name, options) {
        return this.loadCache(name, options);
    }
    loadWithdrawalCache(name, options) {
        return this.loadCache(name, options);
    }
    loadCache(name, options) {
        let constructor, cache = super.loadCache(name);
        if (!cache) {
            const regexp = /([A-Za-z]+)([0-9]+)([A-Z]+)([0-9.]+)/;
            const matches = name.match(regexp);
            if (!matches || matches.length !== 5)
                throw sdk_utils_1.ErrorUtils.getError(`Core.loadCache: name ${name} has wrong format for cache`);
            if (matches[1] === 'Deposits') {
                constructor = (name, options) => new DepositCache(name, options);
            }
            else if (matches[1] === 'Withdrawals') {
                constructor = (name, options) => new WithdrawalCache(name, options);
            }
            else
                throw sdk_utils_1.ErrorUtils.getError(`Core.loadCache: there exists no such event ${matches[1]}`);
            cache = constructor(name, options);
            this.caches.set(name, cache);
        }
        return cache;
    }
    loadInstance(chainId, token, denomination) {
        token = token.toLowerCase();
        return sdk_chain_1.Contracts.getInstance('' + chainId, token, '' + denomination, this.chain.provider);
    }
}
exports.Core = Core;
//# sourceMappingURL=index.js.map